/*****************************************************/
/* about.c                                           */
/* -- About message box.                             */
/*****************************************************/
#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <string.h>
#include "about.h"

#define cidCredits   1000   /* Credits button id. */
#define cidProd      1001   /* Product name id. */
#define cbdrGrid     4      /* Cell in border units. */
#define clrCredit    RGB (0xff, 0x00, 0x00)
#define dtimScroll   33     /* ms per scan line. */
#define cpleAbout    48     /* size of palette. */

/* Round w up to the nearest multiple of dw. */
#define WRoundUp(w, dw) ((((w) + (dw) - 1) / (dw)) * (dw))

typedef struct
	{
	int     x, y;   /* Origin. */
	int     dx, dy; /* Size. */
	WORD    cid;    /* Control ID. */
	LPSTR   lpsz;   /* Title. */
	} CTI;          /* ConTrol Info. */

typedef struct
	{
	/* Control info--must come first. */
	CTI ctiBmp, ctiProd, ctiCopy, ctiMemT, ctiMem,
		ctiProcT, ctiProc, ctiOk, ctiCred;
	int dx, dy;     /* Client area. */
	} ABM;          /* About Box metrics. */

typedef struct
	{
	WORD            wVersion;
	WORD            cple;
	PALETTEENTRY    rgple[cpleAbout];
	} ABL;  /* About Box Palette. */

char szAboutClass[] = "AboutBoxClass";
int  dxGrid, dyGrid;    /* Grid unit. */

BOOL FCalcSzCti (HDC, CTI *, LPSTR);
LRESULT APIENTRY
  LwAboutWndProc (HWND, UINT, WPARAM, LPARAM);
void EnsureWidthAbm (ABM *, LPSTR, HDC);
BOOL FCreateCti (HWND, CTI *);
void InitAbl (ABL *);
void RollCredits (HWND);
void RotatePalette (HPALETTE, ABL *);
LPSTR GetProcessorType (LPSTR, DWORD);

int
FAbout (HINSTANCE hins, HWND hwnd, HICON hicn,
  LPSTR lpszProd, HFONT hfntProd, LPSTR lpszCopyright,
  LPSTR lpszCredits, LPSTR lpszTitle)
/*****************************************************/
/* -- Display an "About" message box.                */
/* -- hins          : Application's instance handle. */
/* -- hwnd          : Owner window (optional).       */
/* -- hicn          : Icon to display (optional).    */
/* -- lpszProd      : Name of product (optional).    */
/* -- hfntProd      : Product name font (optional).  */
/* -- lpszCopyright : Copyright notice (optional).   */
/* -- lpszCredits   : Credit string (optional).      */
/* -- Return false for failure.                      */
/*****************************************************/
    {
    ABM         abm;    /* Control info. */
    int         icti;   /* Control index. */
    HDC         hdc;    /* For text info. */
	int         dyText  = 0;    /* Text height. */
    WNDCLASS    wcs;    /* About window class. */
    HWND        hwndAbout;  /* About window. */
    RECT        rectOwner, rectAbout;
    int         x, y, dx, dy;
    int         dxScreen, dyScreen;
	char        szMem [20], szProc [20];
	MSG         msg;
	TEXTMETRIC  txm;
	MEMORYSTATUS memStat;
	SYSTEM_INFO systemInfo;

	/* Register a class for the main window if there */
	/* isn't one yet. */
	if (GetClassInfo (hins, szAboutClass, &wcs) == 0)
		{
		wcs.style = 0;
		wcs.lpfnWndProc = LwAboutWndProc;
		wcs.cbClsExtra = 0;
		wcs.cbWndExtra = sizeof(HICON) + sizeof(LPSTR);
		wcs.hInstance = hins;
		wcs.hIcon = NULL;
		wcs.hCursor = NULL;
		wcs.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
		wcs.lpszMenuName = NULL;
		wcs.lpszClassName = szAboutClass;
		if (!RegisterClass (&wcs))
			return FALSE;
		}

	if ((hdc = GetDC (NULL)) == NULL)
		return FALSE;

	/* Place items.  Start with white space at edge. */
	memset (&abm, 0, sizeof abm);
	dxGrid = cbdrGrid * GetSystemMetrics (SM_CXBORDER);
	dyGrid = cbdrGrid * GetSystemMetrics (SM_CYBORDER);
	abm.dx = dxGrid;
	abm.dy = dyGrid;

    /* Place bitmap in upper left. */
    abm.ctiBmp.x = abm.dx;
    abm.ctiBmp.y = abm.dy;
    if (hicn != NULL)
        {
        abm.ctiBmp.dx =
          GetSystemMetrics (SM_CXICON) + dxGrid;
        abm.ctiBmp.dy =
          GetSystemMetrics (SM_CYICON) + dyGrid;
        abm.dx += abm.ctiBmp.dx;
        abm.dy += abm.ctiBmp.dy;
        }

    /* Place product name beside bitmap. */
    abm.ctiProd.x = abm.ctiBmp.x + abm.ctiBmp.dx;
    abm.ctiProd.y = abm.ctiBmp.y;
    if (lpszProd == NULL)
		lpszProd = "";  /* For wsprintf() call. */
    else
        {
        HFONT   hfntSav;

		if (hfntProd != NULL)
			hfntSav = SelectObject (hdc, hfntProd);
        if (FCalcSzCti (hdc, &abm.ctiProd, lpszProd))
            {
			abm.dx = abm.ctiProd.x + abm.ctiProd.dx;
			abm.dy = max (abm.dy, abm.ctiProd.dy);
			abm.ctiProd.cid = cidProd;
			}
		SelectObject (hdc, hfntSav);
		}

	/* Place copyright string below product name. */
	abm.ctiCopy.x = abm.ctiProd.x;
	abm.ctiCopy.y = abm.ctiProd.y + abm.ctiProd.dy;
	if (FCalcSzCti (hdc, &abm.ctiCopy, lpszCopyright))
		{
		int dyT;

		if (abm.ctiCopy.dx > abm.ctiProd.dx)
			abm.dx = abm.ctiCopy.x + abm.ctiCopy.dx;
		if ((dyT = abm.ctiCopy.y + abm.ctiCopy.dy) >
		  abm.dy)
			abm.dy = dyT;
		}

	// get system info
	memStat.dwLength = sizeof (MEMORYSTATUS);
	GlobalMemoryStatus (&memStat);
	GetSystemInfo (&systemInfo);
	/* Format strings and make sure there is room to */
	/* show them. */
	wsprintf (szMem, "%ld KB",
		memStat.dwTotalPhys / 1024);
	wsprintf (szProc, "%s",
		GetProcessorType (szProc, systemInfo.dwProcessorType));
	abm.ctiMemT.lpsz = "Physical Memory:";
	abm.ctiMem.lpsz = szMem;
	abm.ctiProcT.lpsz = "Processor Type:";
	abm.ctiProc.lpsz = szProc;
	EnsureWidthAbm (&abm, abm.ctiMemT.lpsz, hdc);
	EnsureWidthAbm (&abm, abm.ctiMem.lpsz, hdc);
	EnsureWidthAbm (&abm, abm.ctiProcT.lpsz, hdc);
	EnsureWidthAbm (&abm, abm.ctiProc.lpsz, hdc);

	/* Make sure client area is wide enough to */
	/* display credits. */
	dxScreen = GetSystemMetrics (SM_CXSCREEN);
	dyScreen = GetSystemMetrics (SM_CYSCREEN);
	if (lpszCredits != NULL)
		{
		RECT    rectT;
		int     dxT;

		rectT.left = rectT.bottom = rectT.top = 0;
		rectT.right = dxScreen;
		if (DrawText (hdc, lpszCredits, -1, &rectT,
		  DT_CALCRECT | DT_EXPANDTABS | DT_NOPREFIX) !=
            0 &&
          (dxT = rectT.right - rectT.left) > abm.dx)
            abm.dx = dxT;
        }

    /* Place memory title and display. */
    abm.ctiMemT.x = dxGrid;
    GetTextMetrics (hdc, &txm);
    dyText = WRoundUp (txm.tmHeight, dyGrid);
    abm.ctiMemT.y = abm.dy + dyGrid;
    abm.ctiMemT.dx = (abm.dx - dxGrid) / 2;
    abm.ctiMemT.dy = dyText + dyGrid;

    abm.ctiMem.x = abm.ctiMemT.x + abm.ctiMemT.dx;
    abm.ctiMem.y = abm.ctiMemT.y;
    abm.ctiMem.dx = abm.ctiMemT.dx;
    abm.ctiMem.dy = abm.ctiMemT.dy;

    /* Place resource title and display. */
	abm.ctiProcT.x = abm.ctiMemT.x;
    abm.ctiProcT.y = abm.ctiMemT.y + abm.ctiMemT.dy;
    abm.ctiProcT.dx = abm.ctiMemT.dx;
    abm.ctiProcT.dy = abm.ctiMemT.dy;

    abm.ctiProc.x = abm.ctiMem.x;
    abm.ctiProc.y = abm.ctiProcT.y;
    abm.ctiProc.dx = abm.ctiMem.dx;
	abm.ctiProc.dy = abm.ctiProcT.dy;
    abm.dy = abm.ctiProc.y + abm.ctiProc.dy;
    ReleaseDC (NULL, hdc);

    /* Place OK and Credits buttons at bottom. */
    abm.ctiOk.dx = (abm.dx - dxGrid) / 2;
    abm.ctiOk.x = lpszCredits == NULL ?
      (abm.dx - abm.ctiOk.dx - dxGrid) / 2 : dxGrid;
    abm.ctiOk.y = abm.dy;
    abm.ctiOk.dy = dyText * 3 / 2 + dyGrid;
    abm.ctiOk.cid = IDOK;
    abm.ctiOk.lpsz = "&OK";
    abm.dy = abm.ctiOk.y + abm.ctiOk.dy;

    if (lpszCredits != NULL)
        {
        abm.ctiCred.x = abm.ctiOk.x + abm.ctiOk.dx;
        abm.ctiCred.y = abm.ctiOk.y;
        abm.ctiCred.dx = abm.ctiOk.dx;
        abm.ctiCred.dy = abm.ctiOk.dy;
        abm.ctiCred.cid = cidCredits;
        abm.ctiCred.lpsz = "&Credits";
        }

    /* Center about window over owner. */
    rectAbout.left = rectAbout.top = 0;
    rectAbout.right = abm.dx;
    rectAbout.bottom = abm.dy;
    AdjustWindowRectEx (&rectAbout, WS_CAPTION |
      WS_SYSMENU, FALSE, WS_EX_DLGMODALFRAME);
    dx = rectAbout.right - rectAbout.left;
    dy = rectAbout.bottom - rectAbout.top;
    if (hwnd == NULL)
        hwnd = GetDesktopWindow ();
    GetWindowRect (hwnd, &rectOwner);
    x = ((rectOwner.right + rectOwner.left) - dx) / 2;
    y = ((rectOwner.bottom + rectOwner.top) - dy) / 2;
    if (x < 0)
        x = 0;
    else if (x + dx > dxScreen)
        x = dxScreen - dx;
    if (y < 0)
        y = 0;
    else if (y + dy > dyScreen)
        y = dyScreen - dy;

    /* Create about window. */
    if ((hwndAbout = CreateWindowEx (
	  WS_EX_DLGMODALFRAME, szAboutClass, lpszTitle,
      WS_POPUP | WS_CAPTION | WS_SYSMENU, x, y, dx, dy,
      hwnd, NULL, hins, NULL)) == NULL)
        return FALSE;
    SetWindowWord (hwndAbout, 0, (WORD)hicn);
    SetWindowLong (hwndAbout, sizeof hicn,
      (LONG)lpszCredits);

    /* Create controls. */
    for (icti = 1; icti < 9; icti++)
        if (!FCreateCti (hwndAbout, (CTI *)&abm + icti))
            {
            DestroyWindow (hwndAbout);
            return FALSE;
            }

    if (abm.ctiProd.lpsz != NULL && hfntProd != NULL)
        SendDlgItemMessage (hwndAbout, cidProd,
          WM_SETFONT, (WPARAM)hfntProd, FALSE);

    ShowWindow (hwndAbout, SW_SHOWNORMAL);
    while (GetMessage (&msg, NULL, 0, 0) &&
      IsWindow (hwndAbout))
        if (!IsDialogMessage (hwndAbout, &msg))
            {
            TranslateMessage (&msg);
            DispatchMessage (&msg);
            }
    return TRUE;
    }

BOOL
FCalcSzCti (HDC hdc, CTI * pcti, LPSTR lpsz)
/*****************************************************/
/* -- Calculate the bounding box for a string.       */
/*****************************************************/
    {
    RECT    rect;

    rect.right = GetSystemMetrics (SM_CXSCREEN) / 2;
    rect.top = rect.left = rect.bottom = 0;
    if (DrawText (hdc, lpsz, -1, &rect, DT_CALCRECT |
      DT_EXPANDTABS | DT_NOPREFIX | DT_WORDBREAK) == 0)
        return FALSE;   // Give up.
    
    pcti->dx = dxGrid +
      WRoundUp (rect.right - rect.left, dxGrid);
    pcti->dy = dyGrid +
      WRoundUp (rect.bottom - rect.top, dyGrid);
    pcti->lpsz = lpsz;
    return TRUE;
    }

void
EnsureWidthAbm (ABM * pabm, LPSTR lpsz, HDC hdc)
/*****************************************************/
/* -- Ensure about box client area can fit two of    */
/*    the given string, side by side.                */
/*****************************************************/
    {
    SIZE    dpt;
    int     dx;

    if (GetTextExtentPoint (hdc, lpsz, lstrlen (lpsz),
      (LPSIZE)&dpt) == 0)
        return;

    dx = 2 * WRoundUp (dpt.cx, dxGrid) + 3 * dxGrid;
    if (dx > pabm->dx)
        pabm->dx = dx;
    }

BOOL
FCreateCti (HWND hwnd, CTI * pcti)
/*****************************************************/
/* -- Create a control.                              */
/* -- Return FALSE for failure.                      */
/*****************************************************/
    {
    DWORD   ulwStyle = WS_CHILD | WS_VISIBLE;
    char *  pszClass = "static";

    if (pcti->cid == 0 && pcti->lpsz == NULL)
        return TRUE;

    if (pcti->cid == IDOK)
        {
        ulwStyle |= WS_TABSTOP | BS_DEFPUSHBUTTON;
        pszClass = "Button";
        }
    else if (pcti->cid == cidCredits)
        {
        ulwStyle |= WS_TABSTOP | BS_PUSHBUTTON;
        pszClass = "Button";
        }
    return CreateWindow (pszClass, pcti->lpsz, ulwStyle,
      pcti->x, pcti->y,
      pcti->dx - dxGrid, pcti->dy - dyGrid,
      hwnd, (HMENU)pcti->cid, GetWindowInstance (hwnd),
      NULL) != NULL;
    }

LRESULT CALLBACK
LwAboutWndProc (HWND hwnd, UINT wm, WPARAM wParam,
  LPARAM lwParam)
/*****************************************************/
/* -- About box main window procedure.               */
/*****************************************************/
    {
    HWND    hwndOwner;

    switch (wm)
        {
    default:
        break;

    case WM_CREATE:
        {
        HMENU   hmnu;

        if ((hwndOwner = GetWindow (hwnd, GW_OWNER)) !=
          NULL)
            EnableWindow (hwndOwner, FALSE);
        hmnu = GetSystemMenu (hwnd, FALSE);
        DeleteMenu (hmnu, SC_TASKLIST, MF_BYCOMMAND);
        DeleteMenu (hmnu, SC_MAXIMIZE, MF_BYCOMMAND);
        DeleteMenu (hmnu, SC_MINIMIZE, MF_BYCOMMAND);
        DeleteMenu (hmnu, SC_SIZE, MF_BYCOMMAND);
        DeleteMenu (hmnu, SC_RESTORE, MF_BYCOMMAND);
        DeleteMenu (hmnu, 3, MF_BYPOSITION);
        DeleteMenu (hmnu, 1, MF_BYPOSITION);
        }
        break;

    case WM_CLOSE:
        if ((hwndOwner = GetWindow (hwnd, GW_OWNER)) !=
          NULL)
            EnableWindow (hwndOwner, TRUE);
        break;

    case WM_ACTIVATE:
        if (wParam != 0)
            {
            LRESULT lw;

            lw =
              DefWindowProc (hwnd, wm, wParam, lwParam);
            SetFocus (GetDlgItem (hwnd, IDOK));
            return lw;
            }
        break;

    case WM_PAINT:
        {
        PAINTSTRUCT wps;

        BeginPaint (hwnd, &wps);
        DrawIcon (wps.hdc, dxGrid, dyGrid,
          (HICON)GetWindowWord (hwnd, 0));
        EndPaint (hwnd, &wps);
        }
        return 0;

    case WM_COMMAND:
        switch (wParam)
            {
        default:
            break;

        case IDOK:
        case IDCANCEL:
            PostMessage (hwnd, WM_CLOSE, 0, 0);
            return 0;

        case cidCredits:
			RollCredits (hwnd);
			return 0;
			}
		break;
		}

	return DefWindowProc (hwnd, wm, wParam, lwParam);
	}

void
RollCredits (HWND hwnd)
/*****************************************************/
/* -- Scroll the credits just like in the movies.    */
/*****************************************************/
    {
	HDC         hdc     = NULL;
	HDC         hdcMem  = NULL;
    HBITMAP     hbmp    = NULL;
    HBITMAP     hbmpSav = NULL;
    HPALETTE    hpal    = NULL;
    RECT        rectClient, rectMem;
    LPSTR       lpsz;
    int         dx, dy, dyScroll;
    int         x, y, yBottom;
    DWORD       tim;
    ABL         abl;

    if ((hdc = GetDC (hwnd)) == NULL)
        return;

    if ((hdcMem = CreateCompatibleDC (hdc)) == NULL)
        goto LCleanup;

    /* Make a bitmap big enough for the credits. */
    lpsz = (LPSTR)GetWindowLong (hwnd, sizeof(HBITMAP));
	GetClientRect (hwnd, &rectClient);
	rectMem = rectClient;
    if (DrawText (hdc, lpsz, -1, &rectMem,
      DT_CALCRECT | DT_CENTER | DT_NOPREFIX |
      DT_WORDBREAK | DT_EXPANDTABS) == 0)
        goto LCleanup;

    /* Create a monochrome bitmap. */
    dx = rectMem.right - rectMem.left;
    dy = rectMem.bottom - rectMem.top;
    if ((hbmp = CreateCompatibleBitmap (hdcMem, dx, dy))
      == NULL)
        goto LCleanup;
    hbmpSav = SelectObject (hdcMem, hbmp);
    PatBlt (hdcMem, 0, 0, dx, dy, WHITENESS);
	DrawText (hdcMem, lpsz, -1, &rectMem, DT_CENTER |
      DT_EXPANDTABS | DT_NOPREFIX | DT_WORDBREAK);

    if (GetDeviceCaps (hdc, RASTERCAPS) & RC_PALETTE)
        {
        InitAbl (&abl);
		if ((hpal = CreatePalette ((LPLOGPALETTE)&abl))
          != NULL)
            {
            SelectPalette (hdc, hpal, FALSE);
            RealizePalette (hdc);
            }
        }

    if (hpal == NULL)
        SetTextColor (hdc, clrCredit);

    SetBkColor (hdc, 0x00000000);
    PatBlt (hdc, 0, 0, rectClient.right,
      rectClient.bottom, BLACKNESS);

    x = (rectClient.right - dx) / 2;
    dyScroll = dy + rectClient.bottom;
    yBottom = rectClient.bottom - 1;
    tim = timeGetTime ();
    for (y = 0; y < dyScroll; y++)
        {
        DWORD   timT;
        MSG     msg;

        if (PeekMessage (&msg, hwnd, WM_KEYDOWN,
              WM_KEYDOWN, PM_REMOVE) ||
          PeekMessage (&msg, hwnd, WM_LBUTTONDOWN,
              WM_LBUTTONDOWN, PM_REMOVE))
            break;

        if (hpal != NULL)
            {
            SetTextColor (hdc,
              PALETTEINDEX (y % cpleAbout));
            RotatePalette (hpal, &abl);
            }
        ScrollDC (hdc, 0, -1, &rectClient, NULL, NULL,
          NULL);
        if (y < dy)
            BitBlt (hdc, x, yBottom, rectClient.right,
              1, hdcMem, 0, y, SRCCOPY);
        while ((timT = timeGetTime ()) - tim <
          dtimScroll)
            ;
        tim = timT;
        }
    InvalidateRect (hwnd, NULL, TRUE);

LCleanup:
    if (hbmpSav != NULL)
        SelectObject (hdcMem, hbmpSav);
    if (hbmp != NULL)
        DeleteObject (hbmp);
    if (hdcMem != NULL)
        DeleteDC (hdcMem);
    if (hdc != NULL)
		ReleaseDC (hwnd, hdc);
    if (hpal != NULL)
        DeleteObject (hpal);
    }

void
RotatePalette (HPALETTE hpal, ABL * pabl)
/*****************************************************/
/* -- Rotate the palette entries up one.             */
/*****************************************************/
    {
    PALETTEENTRY    ple;

    ple = pabl->rgple[0];
    memmove (pabl->rgple, pabl->rgple + 1,
      (cpleAbout - 1) * sizeof pabl->rgple[0]);
    pabl->rgple[cpleAbout - 1] = ple;
    AnimatePalette (hpal, 0, cpleAbout, pabl->rgple);
    }

void
InitAbl (ABL * pabl)
/*****************************************************/
/* -- Initialize the palette with a "sawtooth".      */
/*****************************************************/
    {
    int iple;
    int w   = 0x7f - ((cpleAbout / 4) * 0x20);

    pabl->wVersion = 0x0300;
    pabl->cple = cpleAbout;
    for (iple = 0; iple < cpleAbout; iple++)
        {
        BYTE    b;

        w += iple < cpleAbout / 2 ? 0x20 : -0x20;
		b = w < 0 ? 0 : (w > 0xff ? 0xff : (BYTE)w);
        pabl->rgple[iple].peRed = b;
        pabl->rgple[(iple + cpleAbout / 3) %
          cpleAbout].peGreen = b;
        pabl->rgple[(iple + 2 * cpleAbout / 3) %
          cpleAbout].peBlue = b;
        pabl->rgple[iple].peFlags = PC_RESERVED;
        }
	}

LPSTR
GetProcessorType (LPSTR szProcType, DWORD procMask)
{
	switch (procMask)
	{
		case PROCESSOR_INTEL_386:
			lstrcpy (szProcType, "386"); break;
		case PROCESSOR_INTEL_486:
			lstrcpy (szProcType, "486"); break;
		case PROCESSOR_INTEL_PENTIUM:
			lstrcpy (szProcType, "PENTIUM"); break;
		case PROCESSOR_INTEL_860:
			lstrcpy (szProcType, "Intel 860"); break;
		case PROCESSOR_MIPS_R2000:
			lstrcpy (szProcType, "Mips R2000"); break;
		case PROCESSOR_MIPS_R3000:
			lstrcpy (szProcType, "Mips R3000"); break;
		case PROCESSOR_MIPS_R4000:
			lstrcpy (szProcType, "Mips R4000"); break;
		case PROCESSOR_ALPHA_21064:
			lstrcpy (szProcType, "Alpha 21064"); break;
		default:
			lstrcpy (szProcType, "Unknown"); break;
	}
	return szProcType;
}

