//	magic.c - list all serial port I/O addresses, types & IRQs
//	Mike Dawdy, Compuserve ID 76275,700

// reports the ISR and IRR (of both PICs if it is an AT class machine)
// ISR1 is the IRQ as a decimal number, if the IRQ is between 0 and 7
// ISR2 is the IRQ as a decimal number, if the IRQ is between 8 and 15
// IRQ 2 on an AT class machine always generates IRQ 9 also
// IRQs 8 - 15 on an AT class machine always generate IRQ 2 also
// no bits set to 1 is prints as --
// a single bit set to 1 is printed in decimal as the IRQ number
// multiple bits set to 1 are printed in coded hexadecimal notation as 0xnn
// timeout waiting for interrupt is prints as -- four times

#include <stdio.h>
#include "uart.h"

#define SKIPS 5

void main(void)
{
int irq, i;
unsigned count, io_addr, p, sys_conf;
unsigned long pic;
unsigned char temp;
unsigned char ten_bytes[10];
unsigned char *type[]={"Unknown","INS8250-B","NS16450","NS16550","NS16550A"};
unsigned skip_addresses[SKIPS] = {0x1f0, 0x1f8, 0x320, 0x328, 0x3f0};
unsigned ps2_uarts[] = {0x3220, 0x3228, 0x4220, 0x4228, 0x5220, 0x5228};

printf("I/O Address   Type    Attempts\tISR1 IRR1 ISR2 IRR2");
sys_conf = sys_config() & 0xff02;	// sys_conf == 0x0002 => IBM PS/2

for (io_addr=0x0100, p=0; io_addr<0x0400 || sys_conf==0x0002; io_addr+=0x0008)
 {
 if (io_addr >= 0x0400)
    if (p < 6) io_addr = ps2_uarts[p++];
    else break;

 for (i = 0; i < SKIPS; i++)
  if (io_addr == skip_addresses[i]) break;
 if (i != SKIPS) continue;

 if (uart_check(io_addr) == 0)
  {
  uart_registers(io_addr, ten_bytes);	// save state of UART
  printf("\n    %x     %s\t",io_addr,type[uart_type(io_addr)]);
  i = 1;
  do
   pic = uart_irq(io_addr);
   // pic: 32 bit doubleword, |ISR1|IRR1|ISR2|IRR2|, 8 bits each field
  while ((pic == 0) && (++i < 10));
  printf(" %02u\t", i);
  for (i = 24; i >= 0; i -= 8)	// isolate each 8-bit field of pic
   {
   temp = (char)(pic >> i);
   for (irq = -1, count = 0; temp != 0; temp >>= 1, ++irq)
      count += temp & 0x01;
   if (i < 16) irq += 8;	// it belongs to 2nd 8259A PIC
   if (count == 0) printf(" --  ");
   else
   if (count != 1) printf("%#04x ", (pic >> i) & 0x000000ff);
   else
   printf(" %02u  ", irq);
   }
  uart_restore(io_addr, ten_bytes);	// restore state of UART
  }
 }
}
