//	uart_chk.c - determine if given I/O address is a UART
//	Mike Dawdy, Compuserve ID 76275,700

//	Performs 7 read-only tests to determine if a given I/O
//	address is a UART.  Most non-UART addresses fail test 1.
//	A few fail at test 2 or 3.  Rarely are the rest required.

//	Returns 0 if a UART is found at the given I/O address.
//	Otherwise returns an unsigned, with the low byte set to
//	the test number {1, 2, ..., 7} that failed, and the
//	high byte set to the content of the register that
//	caused the failure.

#include  <dos.h>
#include  "uart.h"

#define   FOUND     0

unsigned uart_check(unsigned addr)
{
unsigned reg;
// certain bits must always be zero
if (((reg = inp(addr+IER)) & 0xf0) != 0x00) return((reg << 8) | 1);
if (((reg = inp(addr+IIR)) & 0x30) != 0x00) return((reg << 8) | 2);
if (((reg = inp(addr+MCR)) & 0xe0) != 0x00) return((reg << 8) | 3);

// reading LSR and TRB forces certain LSR bits to zero
inp(addr+LSR);			// clear OV,PA,FR,BR,RF
inp(addr+TRB);			// clear data ready
if (((reg = inp(addr+LSR)) & 0x9f) != 0x00) return((reg << 8) | 4);

// reading MSR forces certain MSR bits to zero
inp(addr+MSR);			// clear CTS, DSR, RI, DCD
if (((reg = inp(addr+MSR)) & 0x0f) != 0x00) return((reg << 8) | 5);

// IIR bit 0 should be 1, bits 4 & 5 should be 0
if (((reg = inp(addr+IIR)) & 0x31) != 0x01) return((reg << 8) | 6);

// LSR bits 5 and 6 should be 1
if (((reg = inp(addr+LSR)) & 0x60) != 0x60) return((reg << 8) | 7);

return FOUND;
}

